--[[

$Revision: 29 $

(C) Copyright 2009 Bethink (bethink at naef dot com)
See LICENSE.txt for license terms.

]]


----------------------------------------------------------------------
-- Initialization

local L = LibStub("AceLocale-3.0"):GetLocale("BossNotes")


----------------------------------------------------------------------
-- Constants

BOSS_NOTES_SOURCE_HEIGHT = 16
BOSS_NOTES_SOURCE_TEXT_WIDTH = 275
BOSS_NOTES_SOURCE_DISPLAYED_COUNT = 8
BOSS_NOTES_REFRESH_COOLDOWN = 10

-- Chats
local CHATS = {
	{
		text = L["RAID"],
		value = "RAID"
	},
	{
		text = L["PARTY"],
		value = "PARTY"
	},
	{
		text = L["SAY"],
		value = "SAY"
	},
	{
		text = L["YELL"],
		value = "YELL"
	},
	{
		text = L["WHISPER"],
		value = "WHISPER"
	},
	{
		text = L["GUILD"],
		value = "GUILD"
	},
	{
		text = L["OFFICER"],
		value = "OFFICER"
	}
}

	
---------------------------------------------------------------------
-- Subsystem

-- Initializes the Boss Notes frame subsystem
function BossNotes:InitializeBossNotesFrame ()
	-- Update
	self:UpdateInstances()
	
	-- Initialize chats
	self:InitializeChats()
end

-- Enable the Boss Notes frame subsystem
function BossNotes:EnableBossNotesFrame ()
	self:RegisterEvent("UNIT_PORTRAIT_UPDATE")
	self:RegisterMessage("BOSS_NOTES_UPDATE_SOURCES")
	self:RegisterMessage("BOSS_NOTES_ACTIVATION")
end

-- Disables the Boss Notes frame subsystem
function BossNotes:DisableBossNotesFrame ()
end


----------------------------------------------------------------------
-- APIs

-- Returns the current selection
function BossNotes:GetSelection ()
	local selectedInstance = self.selectedInstance
	local selectedEncounter = self.selectedEncounter
	local selectedSource = self.selectedSource
	return selectedInstance, selectedEncounter, selectedSource
end


----------------------------------------------------------------------
-- Event handlers

-- OnLoad handler
function BossNotes:OnLoadBossNotesFrame (frame)
	-- Make the Boss Notes frame pushable
	BossNotesFrame:SetAttribute("UIPanelLayout-defined", true)
	BossNotesFrame:SetAttribute("UIPanelLayout-enabled", true)
	BossNotesFrame:SetAttribute("UIPanelLayout-area", "left")
	BossNotesFrame:SetAttribute("UIPanelLayout-pushable", 5)
	BossNotesFrame:SetAttribute("UIPanelLayout-whileDead", true)

	-- Translations
	BossNotesFrameTitleText:SetText(L["BOSS_NOTES_FRAME_TITLE"])
	BossNotesRefreshButton:SetText(L["REFRESH"])
end

-- OnShow handler
function BossNotes:OnShowBossNotesFrame (frame)
	PlaySound("igCharacterInfoOpen")
	SetPortraitTexture(BossNotesFramePortrait, "player")
	self:UpdateSources()
end

-- OnHide handler
function BossNotes:OnHideBossNotesFrame (frame)
	PlaySound("igCharacterInfoClose")
	if BossNotesEditBoxPopup:IsShown() then
		BossNotesEditBoxPopup:Hide()
	end
end

-- Handles the clicking of the note
function BossNotes:OnClickNoteScrollFrame (frame)
	-- Only open editor if the note is editable
	if not self.canEditNote then
		return
	end
	
	-- Store edit key
	self.editKey = self.selectedKey

	-- Set text and note, show editor
	local selectedInstance = self.selectedInstance
	local selectedEncounter = self.selectedEncounter
	local source = self.selectedSource
	BossNotesEditBoxPopup.text:SetText(string.format(L["UI_SOURCE_ON_CONTEXT"],
			source.color.r * 255, source.color.g * 255, source.color.b * 255,
			source.name, self:GetContextText(selectedInstance, selectedEncounter)))
	BossNotesEditBoxPopup.editBox:SetText(source.note or "")
	BossNotesEditBoxPopup:Show()
end

-- Handles the entering into a hyperlink
function BossNotes:OnHyperlinkEnterNote (frame, link, text)
	GameTooltip:SetOwner(frame, "ANCHOR_BOTTOMRIGHT", 50, 0)
	GameTooltip:SetHyperlink(link)
	GameTooltip:Show()
end

-- Handles the leaving of a hyperllink
function BossNotes:OnHyperlinkLeaveNote (frame, link, text)
	GameTooltip:Hide()
end

-- Handles the clicking of a hyperllink
function BossNotes:OnHyperlinkClickNote (frame, link, text, button)
	if IsModifiedClick("CHATLINK") then
		ChatEdit_InsertLink(text)
	end
end

-- Handles the leaving of a hyperllink
function BossNotes:OnHyperlinkLeaveNote (frame, link, text)
	GameTooltip:Hide()
end
-- Handles the clicking of the refresh button
function BossNotes:OnClickRefresh (frame)
	-- Force a source request on the network
	self:RequestSources(self.selectedInstance, self.selectedEncounter, true)
	
	-- Disable the button and set a timer
	frame:Disable()
	self:ScheduleTimer("EnableRefresh", BOSS_NOTES_REFRESH_COOLDOWN)
end

-- Enables the refresh button
function BossNotes:EnableRefresh ()
	BossNotesRefreshButton:Enable()
end

-- Handles the UNIT_PORTRAIT_UPDATE event
function BossNotes:UNIT_PORTRAIT_UPDATE (message, unit)
	if unit == "player" then
		SetPortraitTexture(BossNotesFramePortrait, "player")
	end
end

-- Handles the BOSS_NOTES_UPDATE_SOURCES message from the data subsystem
function BossNotes:BOSS_NOTES_UPDATE_SOURCES (message, keys)
	if keys[self.selectedKey] then
		self:UpdateSources()
	end
end

-- Handles an activation evevent from the data subsystem
function BossNotes:BOSS_NOTES_ACTIVATION (message, instance, encounter)
	-- Update instance
	UIDropDownMenu_SetSelectedValue(BossNotesInstanceDropDown, instance)
	UIDropDownMenu_SetText(BossNotesInstanceDropDown, instance.name)
	self.selectedInstance = instance
	self:UpdateEncounters(true)
	
	-- Update encounter
	UIDropDownMenu_SetSelectedValue(BossNotesEncounterDropDown, encounter)
	self.selectedEncounter = encounter
	self:UpdateSources()
end


----------------------------------------------------------------------
-- Edit box popup, adopted from Khaos

-- Handles the loading of the multline edit box popup
function BossNotes:OnLoadEditBoxPopup (frame)
	local name = frame:GetName()

	-- Get and store child frames
	frame.text = getglobal(name .. "Text")
	frame.accept = getglobal(name .. "Button1")
	frame.cancel = getglobal(name .. "Button2")
	frame.scrollFrameBackground = getglobal(name .. "ScrollFrameBackground")
	frame.editBox = getglobal(name .. "ScrollFrameEditBox")
	
	-- Accept button
	frame.accept:SetText(ACCEPT)
	frame.accept:SetPoint("TOPRIGHT", frame.scrollFrameBackground, "BOTTOM", -6, -8)
	frame.accept:Show()
	frame.accept:SetScript("OnClick", BossNotes_OnClickEditBoxPopupAccept)
	
	-- Cancel button
	frame.cancel:SetText(CANCEL)
	frame.cancel:SetPoint("LEFT", frame.accept, "RIGHT", 13, 0)
	frame.cancel:Show()
	frame.cancel:SetScript("OnClick", BossNotes_OnClickEditBoxPopupCancel)

	-- Hide unused frames
	getglobal(name .. "AlertIcon"):Hide()
	getglobal(name .. "MoneyInputFrame"):Hide()
end

-- Handles the showing of the edit box popup
function BossNotes:OnShowEditBoxPopup (frame)
	-- Update height
	local textHeight = frame.text:GetHeight()
	local buttonHeight = frame.accept:GetHeight()
	local editBoxHeight = frame.scrollFrameBackground:GetHeight()
	frame:SetHeight(16 + textHeight + 8 + editBoxHeight + 8 + buttonHeight + 36)
	
	-- Set focus
	frame.editBox:SetFocus()
end
	
-- Handles the hiding of the edit box popup
function BossNotes:OnHideEditBoxPopup (frame)
end

-- Handles the clicking of the accept button
function BossNotes_OnClickEditBoxPopupAccept (frame)
	-- Hide editor
	BossNotesEditBoxPopup:Hide()
	
	-- Update note
	local text = frame:GetParent().editBox:GetText()
	if BossNotes:SetPlayerNote(BossNotes.editKey, text) then
		BossNotes:UpdateSources()
	end
end

-- Handles the clicking of the cancel button
function BossNotes_OnClickEditBoxPopupCancel (frame)
	-- Hide editor
	BossNotesEditBoxPopup:Hide()
end


----------------------------------------------------------------------
-- Chats

-- Updates the chat drop down
function BossNotes:InitializeChats ()
	UIDropDownMenu_Initialize(BossNotesChatDropDown, BossNotesChatDropDown_Initialize)
	UIDropDownMenu_SetText(BossNotesChatDropDown, L["SEND_TO"])
end

-- Initializes the chat drop down
function BossNotesChatDropDown_Initialize ()
	-- Add effective chats
	for _, chat in ipairs(CHATS) do
		info = UIDropDownMenu_CreateInfo()
		info.text = chat.text
		info.value = chat.value
		info.func = BossNotes_OnChatDropDownClick
		UIDropDownMenu_AddButton(info)
	end
end

-- Handles the selection of a chat
function BossNotes_OnChatDropDownClick (button)
	-- Get the note
	local note = BossNotes.selectedSource.note
	if not note then
		return
	end
	note = BossNotes:StripHtml(strtrim(note))
	if string.len(note) == 0 then
		return
	end
	
	-- Get the chat to send to
	local chat = button.value
	
	-- Get the destination
	local destination = nil
	if chat == "WHISPER" then
		if not UnitIsPlayer("target") then
			return
		end
		destination = UnitName("target")
	end
	
	-- Send via ChatThrottleLib which is hard-embedded into AceComm-3.0
	ChatThrottleLib:SendChatMessage("NORMAL", "", string.format(L["CHAT_SOURCE_ON_CONTEXT"],
			BossNotes.selectedSource.name, BossNotes:GetContextText(BossNotes.selectedInstance,
			BossNotes.selectedEncounter)), chat, nil, destination)
	for line in string.gmatch(note, "[^\r\n]+") do
		local pos = 1
		line = string.gsub(line, "()%s+()(%S+)()", function (prev, start, word, stop)
			if stop - pos > 60 then
				ChatThrottleLib:SendChatMessage("NORMAL", "", string.sub(line, pos, prev - 1),
						chat, nil, destination)
				pos = start
			end
		end)
		ChatThrottleLib:SendChatMessage("NORMAL", "", string.sub(line, pos),
				chat, nil, destination)
	end
end


----------------------------------------------------------------------
-- Content and selection managment

-- Updates the instances drop down
function BossNotes:UpdateInstances (limitted)
	UIDropDownMenu_Initialize(BossNotesInstanceDropDown, BossNotesInstanceDropDown_Initialize)
	UIDropDownMenu_SetSelectedID(BossNotesInstanceDropDown, 1)
	self.selectedInstance = nil
	if not limitted then
		self:UpdateEncounters()
	end
end

-- Initializes the instances drop down
function BossNotesInstanceDropDown_Initialize ()
	-- Add "All Instances"
	local info = UIDropDownMenu_CreateInfo()
	info.text = L["ALL_INSTANCES"]
	info.func = BossNotes_OnInstanceDropDownClick
	UIDropDownMenu_AddButton(info)

	-- Add effective instances
	local instances = BossNotes:GetInstances()
	for _, instance in ipairs(instances) do 
		info = UIDropDownMenu_CreateInfo()
		info.text = BossNotes:GetContextText(instance, nil)
		info.func = BossNotes_OnInstanceDropDownClick
		info.value = instance
		UIDropDownMenu_AddButton(info)
	end
end

-- Handles the selection of an instance
function BossNotes_OnInstanceDropDownClick (button)
	UIDropDownMenu_SetSelectedID(BossNotesInstanceDropDown, button:GetID())
	BossNotes.selectedInstance = button.value
	if type(BossNotes.selectedInstance) ~= "table" then
		BossNotes.selectedInstance = nil
	end
	BossNotes:UpdateEncounters()
end

-- Updates the encounters drop down
function BossNotes:UpdateEncounters (limitted)
	UIDropDownMenu_Initialize(BossNotesEncounterDropDown, BossNotes_InitializeEncounterDropDown)
	UIDropDownMenu_SetSelectedID(BossNotesEncounterDropDown, 1)
	self.selectedEncounter = nil
	if not limitted then
		self:UpdateSources()
	end
end

-- Initializes the encounters drop down
function BossNotes_InitializeEncounterDropDown ()
	-- Add "All Encounters"
	local info = UIDropDownMenu_CreateInfo()
	info.text = L["ALL_ENCOUNTERS"]
	info.func = BossNotes_OnEncounterDropDownClick
	UIDropDownMenu_AddButton(info)
	
	-- Add effective encounters
	if BossNotes.selectedInstance then
		for _, encounter in ipairs(BossNotes.selectedInstance.encounters) do
			info = UIDropDownMenu_CreateInfo()
			info.text = BossNotes:GetContextText(BossNotes.selectedInstance, encounter)
			info.func = BossNotes_OnEncounterDropDownClick
			info.value = encounter
			UIDropDownMenu_AddButton(info)
		end
	end
end

-- Handles the selection of an encounter
function BossNotes_OnEncounterDropDownClick (button)
	UIDropDownMenu_SetSelectedID(BossNotesEncounterDropDown, button:GetID())
	BossNotes.selectedEncounter = button.value
	if type(BossNotes.selectedEncounter) ~= "table" then
		BossNotes.selectedEncounter = nil
	end
	BossNotes:UpdateSources()
end

-- Update the list of sources
function BossNotes:UpdateSources (limitted)
	-- Request sources (no force)
	self:RequestSources(self.selectedInstance, self.selectedEncounter, false)
	
	-- Update sources
	self.selectedSources = self:GetSources(self.selectedInstance, self.selectedEncounter)
	
	-- Maintain selection
	local oldSelectedSource = self.selectedSource
	self.selectedSource = nil
	if oldSelectedSource then
		for _, source in ipairs(self.selectedSources) do
			if source.name == oldSelectedSource.name then
				self.selectedSource = source
				break
			end
		end
	end
	if not self.selectedSource then
		self.selectedSource = self.selectedSources[1]
	end

	-- Send selection change message
	local oldSelectedKey = self.selectedKey
	self.selectedKey = self:GetContextKey(self.selectedInstance, self.selectedEncounter)
	if self.selectedKey ~= oldSelectedKey then
		self:SendMessage("BOSS_NOTES_SELECTION", self.selectedInstance, self.selectedEncounter)
	end
	
	-- Update source
	if not limitted then
		self:UpdateSourceList()
	end
end

-- Update the source list
function BossNotes:UpdateSourceList (limitted)
	-- Update the scroll framne
	local numSources = #self.selectedSources
	local sourceOffset = FauxScrollFrame_GetOffset(BossNotesSourceListScrollFrame)
	FauxScrollFrame_Update(BossNotesSourceListScrollFrame, numSources, BOSS_NOTES_SOURCE_DISPLAYED_COUNT,
			BOSS_NOTES_SOURCE_HEIGHT, nil, nil, nil, BossNotesSourceHighlightFrame, 293, 316)
	BossNotesSourceHighlightFrame:Hide()
	for i = 1, BOSS_NOTES_SOURCE_DISPLAYED_COUNT do
		local sourceButton = _G["BossNotesSource" .. i]
		local sourceButtonText = _G["BossNotesSource" .. i .. "Text"]
		local sourceIndex = i + sourceOffset
		local source = self.selectedSources[sourceIndex]
		if sourceIndex <= numSources then	
			if BossNotesSourceListScrollFrame:IsShown() then
				sourceButton:SetWidth(293)
			else
				sourceButton:SetWidth(323)
			end
			sourceButtonText:SetTextColor(source.color.r, source.color.g, source.color.b)
			sourceButton:SetText(" " .. source.name)
			sourceButton:SetID(sourceIndex)
			sourceButton:Show()
			if source == self.selectedSource then
				BossNotesSourceHighlightFrame:SetPoint("TOPLEFT", "BossNotesSource" .. i, "TOPLEFT", 2, 0)
				BossNotesSourceHighlightFrame:Show()
				sourceButton:LockHighlight()
				sourceButton.isHighlighted = true
			else
				sourceButton:UnlockHighlight()
				sourceButton.isHighlighted = false
			end
		else
			sourceButton:Hide()
		end
	end
	
	-- Update the source
	if not limitted then
		self:UpdateSource()
	end
end

-- Regular function invoked by the scroll bar
function BossNotesFrame_UpdateSourceList ()
	BossNotes:UpdateSourceList()
end

-- Handles the clicking of a source button
function BossNotes:OnSourceButtonClick (frame, button)
	if button == "LeftButton" then
		self.selectedSource = self.selectedSources[frame:GetID()]
		self:UpdateSourceList()
	end
end

-- Update the source
function BossNotes:UpdateSource (limitted)
	local source = self.selectedSource
	self.canEditNote = source.name == self.playerName
	if source.note then
		BossNotesNote:SetText(source.note)
	else
		if self.canEditNote then
			BossNotesNote:SetText(L["CLICK_TO_EDIT"])
		else
			BossNotesNote:SetText("")
		end
	end
	if source.time and source.clientVersion then
		BossNotesMetaData:SetText(string.format("%s (%s)",
				date("%Y-%m-%d %H:%M", source.time),
				source.clientVersion))
	else
		BossNotesMetaData:SetText("")
	end
end


----------------------------------------------------------------------
-- Support methods

-- Removes HTML from a string
function BossNotes:StripHtml (html)
	if string.sub(html, 1, 1) ~= "<" then
		return html
	end
	return string.gsub(html, "([^<]*)<([^>]+)>", function (content, element)
		if string.sub(element, 1, 1) == "/" or string.sub(element, -1) == "/" then
			return content .. "\n"
		else
			return content
		end
	end)
end
