--[[

$Revision: 29 $

(C) Copyright 2009 Bethink (bethink at naef dot com)
See LICENSE.txt for license terms.

]]


----------------------------------------------------------------------
-- Initialization

local L = LibStub("AceLocale-3.0"):GetLocale("BossNotes")


----------------------------------------------------------------------
-- Constants

-- Cache TTL
BOSS_NOTES_TTL = 300

-- Orders
local ORDER_PLAYER = 1
local ORDER_COLLECTION = 2
local ORDER_COMMUNITY = 3

-- Collection color (rgb = 0x71d5ff)
local COLLECTION_COLOR = { r = 0.444, g = 0.836, b = 1.0 }


---------------------------------------------------------------------
-- Subsystem

-- Initializes the data subsystem
function BossNotes:InitializeData ()
	-- Player name and class
	self.playerName = UnitName("player")
	_, self.playerClass = UnitClass("player")
	
	-- Sources
	self.sources = { }
	
	-- Collections
	self.collections = { }
	
	-- Cache
	self.caches = { }
	
	-- NPC to instance/encounter map
	self:UpdateInstanceEncounters()
end

-- Enables the data subsystem
function BossNotes:EnableData ()
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterMessage("BOSS_NOTES_REQUEST")
	self:RegisterMessage("BOSS_NOTES_RESPONSE")
end

-- Disables the data subsystem
function BossNotes:DisableData  ()
end

----------------------------------------------------------------------
-- APIs

-- Returns the instances
function BossNotes:GetInstances ()
	local instances = { }
	for _, instanceSet in ipairs(BOSS_NOTES_ENCOUNTERS) do
		if self.db.global["instanceSet" .. instanceSet.instanceSetId] then
			for _, instance in ipairs(instanceSet.instances) do
				table.insert(instances, instance)
			end
		end
	end
	return instances
end

-- Requests the sources on the network
function BossNotes:RequestSources (instance, encounter, force)
	-- Request from network if TTL has expired
	local key = self:GetContextKey(instance, encounter)
	local cache = self:GetCache(key)
	local time = time()
	if time - cache.lastRequest > BOSS_NOTES_TTL or force then
		self:SendRequest(key)
	end
end

-- Returns the sources
function BossNotes:GetSources (instance, encounter)
	-- Update sources as needed
	local key = self:GetContextKey(instance, encounter)
	if not self.sources[key] then
		self.sources[key] = { }
		local sources = self.sources[key]
	
		-- Add player
		local source = {
			order = ORDER_PLAYER,
			name = self.playerName,
			color = RAID_CLASS_COLORS[self.playerClass]
		}
		if self.db.char.notes[key] then
			source.note = self.db.char.notes[key].note
			source.time = self.db.char.notes[key].time
			source.clientVersion = self.db.char.notes[key].clientVersion
		end
		table.insert(sources, source)
		
		-- Add from collections
		for name, queryFunc in pairs(self.collections) do
			local note = queryFunc(instance, encounter)
			if note then
				local source = {
					order = ORDER_COLLECTION,
					name = name,
					color = COLLECTION_COLOR,
					note = note
				}
				table.insert(sources, source)
			end
		end

		-- Add from cache
		local cache = self:GetCache(key)
		for name, source in pairs(cache.sources) do
			local source = {
				order = ORDER_COMMUNITY,
				name = name,
				color = RAID_CLASS_COLORS[source.class],
				note = source.note,
				time = source.time,
				clientVersion = source.clientVersion
			}
			table.insert(sources, source)
		end
		
		-- Sort
		table.sort(sources, function (a, b)
			if a.order ~= b.order then
				return a.order < b.order
			end
			local aTime = a.time or 0
			local bTime = b.time or 0
			if aTime ~= bTime then
				return aTime > bTime
			end
			return a.name < b.name
		end)
	end
	
	return self.sources[key]
end

-- Sets the note of the player
function BossNotes:SetPlayerNote (key, note)
	-- Normalize note to nil if empty
	if note then
		note = strtrim(note)
		if string.len(note) == 0 then
			note = nil
		end
	end
	
	-- Set or clear, invalidating sources as needed
	if note and (not self.db.char.notes[key] or note ~= self.db.char.notes[key].note) then
		self.db.char.notes[key] = {
			note = note,
			time = time(),
			clientVersion = GetBuildInfo()
		}
		self:InvalidateContextKey(key, true)
	end
	if not note and self.db.char.notes[key] then
		self.db.char.notes[key] = nil
		self:InvalidateContextKey(key, true)
	end
end

-- Updates the instance/encounter map
function BossNotes:UpdateInstanceEncounters ()
	self.instanceEncounters = { }
	for _, instanceSet in ipairs(BOSS_NOTES_ENCOUNTERS) do
		if self.db.global["instanceSet" .. instanceSet.instanceSetId] then
			for _, instance in ipairs(instanceSet.instances) do
				for _, encounter in ipairs(instance.encounters) do
					local instanceEncounter = {
						instance = instance,
						encounter = encounter
					}
					for _, npcId in ipairs (encounter.npcIds) do
						self.instanceEncounters[npcId] = instanceEncounter
					end
				end
			end
		end
	end
end

-- Returns whether an instance and an encounter for an NPC ID
function  BossNotes:GetInstanceAndEncounter (npcId)
	local instanceEncounter = self.instanceEncounters[npcId]
	if not instanceEncounter then
		return
	end
	return instanceEncounter.instance, instanceEncounter.encounter
end

-- Registers a collection
function BossNotes:RegisterCollection (name, queryFunc)
	self.collections[name] = queryFunc
end

-- Unregisters a collection
function BossNotes:UnregisterCollection (name)
	self.collections[name] = nil
end

-- Returns a key for a context
function BossNotes:GetContextKey (instance, encounter) 
	if instance then
		if encounter then
			return "encounter:" .. tostring(encounter.encounterId)
		else
			return "instance:" .. instance.instanceId
		end
	else
		return "general"
	end
end

-- Returns a text representation for a context
function BossNotes:GetContextText (instance, encounter)
	if instance then
		if encounter then
			return encounter.name
		else
			return instance.name
		end
	else
		return L["GENERAL"]
	end
end

-- Invalidates a context key and schedules an asynchronous UI update
function BossNotes:InvalidateContextKey (key, immediate)
	-- Clear sources
	self.sources[key] = nil
		
	-- Update
	if not self.updateSourcesKeys then
		self.updateSourcesKeys = { }
	end
	self.updateSourcesKeys[key] = true
	if immediate then
		self:SendUpdateSources()
	else
		if not self.updateSourcesTimer then
			self.updateSourcesTimer = self:ScheduleTimer("SendUpdateSources", 1.0)
		end
	end
end

-- Sends the sources update message once the timer has expired
function BossNotes:SendUpdateSources ()
	-- Clean
	self.updateSourcesTimer = nil
	local keys = self.updateSourcesKeys
	self.updateSourcesKeys = nil
	
	-- Send message
	self:SendMessage("BOSS_NOTES_UPDATE_SOURCES", keys)
end

----------------------------------------------------------------------
-- Event handlers

-- Handles the UPDATE_MOUSEOVER_UNIT event
function BossNotes:UPDATE_MOUSEOVER_UNIT ()
	self:CheckUnit("mouseover")
end

-- Handles the PLAYER_TARGET_CHANGED event
function BossNotes:PLAYER_TARGET_CHANGED ()
	self:CheckUnit("target")
end

-- Handles a request
function BossNotes:BOSS_NOTES_REQUEST (message, key, distribution)
	-- Note time of request
	local cache = self:GetCache(key)
	cache.lastRequest = time()
	
	-- Send response
	local note = self.db.char.notes[key]
	if note then
		self:SendResponse(key, self.playerClass, note.note, note.time, note.clientVersion, distribution)
	end
end

-- Handles a response
function BossNotes:BOSS_NOTES_RESPONSE (message, key, name, class, note, time, clientVersion)
	-- Ignore our own response
	if name == self.playerName then
		return
	end
	
	-- Cache response
	local modified = false
	local cache = self:GetCache(key)
	local source = cache.sources[name]
	if not source then
		cache.sources[name] = {
			class = class
		}
		source = cache.sources[name]
		modified = true
	end
	modified = modified or note ~= source.note or time ~= source.time or clientVersion ~= source.clientVersion
	if modified then
		-- Update
		source.note = note
		source.time = time
		source.clientVersion = clientVersion

		-- Invalidate
		self:InvalidateContextKey(key)
	end
end


----------------------------------------------------------------------
-- Support methods

-- Returns the cache for a key
function BossNotes:GetCache (key)
	if not self.caches[key] then
		self.caches[key] = {
			lastRequest = 0,
			sources = { }
		}
	end
	return self.caches[key]
end

-- Checks a unit
function BossNotes:CheckUnit (unit)
	-- Get the GUID
	local guid = UnitGUID(unit)
	
	-- Test whether the unit is a creature
	if guid and string.sub(guid, 5, 5) == "3" then
		-- Get the NPC ID of the unit
		local npcId = tonumber("0x" .. string.sub(guid, 9, 12))
		
		-- Find and process a mapped instance encounter
		local instance, encounter = self:GetInstanceAndEncounter(npcId)
		if instance and encounter then
			self:SendMessage("BOSS_NOTES_ACTIVATION", instance, encounter)
		end
	end
end
